use adblock::engine::Engine;
use adblock::lists::{ParseOptions, RuleTypes};
use anyhow::Context;
use ipnetwork::IpNetwork;
use kuchiki::NodeRef;
use lingua::Language;
use std::collections::{BTreeSet, HashSet};
use std::net::IpAddr;
use tokio::io::{AsyncBufReadExt, AsyncRead, BufReader};

pub async fn load_adblock_engine<R: AsyncRead + Unpin>(
    reader: R,
    rule_types: RuleTypes,
) -> anyhow::Result<Engine> {
    let mut br = BufReader::new(reader);
    let mut rules = Vec::new();
    let mut buf = String::new();
    loop {
        buf.clear();
        if br.read_line(&mut buf).await? == 0 {
            break;
        }
        rules.push(buf.trim().to_owned());
    }
    Ok(Engine::from_rules(
        &rules,
        ParseOptions {
            format: Default::default(),
            include_redirect_urls: false,
            rule_types,
        },
    ))
}

// Relevant:
// https://github.com/brave/adblock-rust/issues/152#issuecomment-771259069

pub struct ExtractedClassesAndIds {
    classes: Vec<String>,
    ids: Vec<String>,
}

pub fn extract_classes_and_ids_from_page(root: &NodeRef) -> ExtractedClassesAndIds {
    let mut class_set = HashSet::new();
    let mut id_set = HashSet::new();

    for node in root.inclusive_descendants() {
        if let Some(element) = node.0.as_element() {
            let attrs = element.attributes.borrow();
            if let Some(id) = attrs.get("id") {
                id_set.insert(id.to_owned());
            }
            if let Some(classes) = attrs.get("class") {
                for class in classes.trim().split_whitespace() {
                    class_set.insert(class.to_owned());
                }
            }
        }
    }

    ExtractedClassesAndIds {
        classes: class_set.into_iter().collect(),
        ids: id_set.into_iter().collect(),
    }
}

pub fn analyse_with_ad_block_cosmetic_filter(
    root: &NodeRef,
    engine: &Engine,
    url: &str,
    remove: bool,
) -> anyhow::Result<bool> {
    let mut matches = 0;

    let url_resources = engine.url_cosmetic_resources(url);
    let specialist_hide_selectors = if !url_resources.generichide {
        let ExtractedClassesAndIds { classes, ids } = extract_classes_and_ids_from_page(root);

        //eprintln!("ID {:#?}", ids);
        //eprintln!("CC {:#?}", classes);

        engine.hidden_class_id_selectors(&classes, &ids, &url_resources.exceptions)
    } else {
        Vec::with_capacity(0)
    };

    //eprintln!("UR {:#?}", url_resources);
    //eprintln!("sHS {:#?}", specialist_hide_selectors);
    //eprintln!("----");

    for rule in itertools::chain(specialist_hide_selectors, url_resources.hide_selectors) {
        if let Ok(result) = root.select(&rule) {
            for ele in result {
                eprintln!("Cosmetic Filter {:?} Matches {:?}", rule, ele);
                matches += 1;
                if remove {
                    ele.as_node().detach();
                }
            }
        } else {
            //eprintln!("(fail)");
        }
    }
    Ok(matches > 0)
}

pub fn guess_document_language(text: &str) -> Option<String> {
    let detector = lingua::LanguageDetectorBuilder::from_all_languages().build();
    detector
        .detect_language_of(text)
        .map(|lang: Language| lang.iso_code_639_1().to_string())
}

// TODO this isn't particularly efficient. Probably want a trie if it's important...
pub struct IpSet {
    ips: BTreeSet<IpNetwork>,
}

impl IpSet {
    pub fn new() -> IpSet {
        IpSet {
            ips: Default::default(),
        }
    }

    pub async fn add_all_from_file<R: AsyncRead + Unpin>(
        &mut self,
        reader: R,
    ) -> anyhow::Result<()> {
        let mut br = BufReader::new(reader);

        let mut buf = String::new();
        loop {
            buf.clear();
            if br.read_line(&mut buf).await? == 0 {
                break;
            }

            let trimmed = buf.trim();

            if trimmed.is_empty() {
                continue;
            }

            let ip_net = trimmed
                .parse::<IpNetwork>()
                .context("Parsing CIDR IP range")?;
            self.add(ip_net);
        }

        Ok(())
    }

    pub fn add(&mut self, network: IpNetwork) {
        // We jump through a couple of hoops to make sure we store the lowest address in the network,
        // since we use that for sorting.
        self.ips
            .insert(IpNetwork::new(network.network(), network.prefix()).unwrap());
    }

    pub fn contains(&self, addr: IpAddr) -> bool {
        let prefix = if addr.is_ipv4() {
            32
        } else {
            assert!(addr.is_ipv6());
            128
        };
        let addr_as_net =
            IpNetwork::new(addr, prefix).expect("Conversion to IpNetwork should be correct");
        for ipnet in self.ips.range(..=addr_as_net).rev().next() {
            if ipnet.contains(addr) {
                return true;
            }
        }

        false
    }
}

#[cfg(test)]
mod test {
    use crate::raking::analysis::IpSet;
    use ipnetwork::IpNetwork;
    use std::net::IpAddr;
    use std::str::FromStr;

    #[test]
    pub fn test_ipset_contains() {
        let mut set = IpSet::new();
        set.add(IpNetwork::from_str("1.2.3.4/16").unwrap());
        set.add(IpNetwork::from_str("1.1.2.3/16").unwrap());
        set.add(IpNetwork::from_str("85.42.36.17/24").unwrap());

        assert!(set.contains(IpAddr::from_str("1.2.42.42").unwrap()));
        assert!(set.contains(IpAddr::from_str("85.42.36.14").unwrap()));
        assert!(!set.contains(IpAddr::from_str("85.42.37.14").unwrap()));
    }
}
