use crate::raking::UrlRaked;
use quickpeep_densedoc::{DenseDocument, DenseTree};
use quickpeep_structs::rake_entries::{RakedReference, ReferenceKind};
use reqwest::Url;
use std::collections::BTreeSet;

pub fn find_references(
    doc: &DenseDocument,
    feeds: &Vec<Url>,
    page_url: &Url,
) -> BTreeSet<RakedReference> {
    let mut refs = BTreeSet::new();

    fn add_link_refs(tree: &Vec<DenseTree>, refs: &mut BTreeSet<RakedReference>, page_url: &Url) {
        for node in tree {
            match node {
                DenseTree::Heading1(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading2(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading3(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading4(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading5(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Heading6(children) => {
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Link {
                    children,
                    href,
                    nofollow,
                } => {
                    if !nofollow {
                        if let Ok(full_url) = page_url.join(&href) {
                            refs.insert(RakedReference {
                                target: clean_url(&full_url).to_string(),
                                kind: ReferenceKind::Link,
                            });
                        }
                    }
                    add_link_refs(children, refs, page_url);
                }
                DenseTree::Image { .. } => {}
                DenseTree::Text(_) => {}
            }
        }
    }

    add_link_refs(&doc.body_content, &mut refs, &page_url);
    add_link_refs(&doc.body_remainder, &mut refs, &page_url);

    for feed in feeds {
        refs.insert(RakedReference {
            target: clean_url(feed).as_str().to_owned(),
            kind: ReferenceKind::HeaderLinkedFeed,
        });
    }

    refs
}

pub fn references_from_urlrakes(
    input: &Vec<UrlRaked>,
    ref_kind: ReferenceKind,
) -> BTreeSet<RakedReference> {
    input
        .iter()
        .map(|url_raked| RakedReference {
            target: url_raked.url.to_string(),
            kind: ref_kind,
        })
        .collect()
}

pub fn clean_url(url: &Url) -> Url {
    let mut url = url.clone();
    url.set_fragment(None);

    url
}

#[cfg(test)]
mod test {
    use crate::raking::references::clean_url;
    use reqwest::Url;
    use std::str::FromStr;

    #[test]
    pub fn test_clean_url() {
        assert_eq!(
            clean_url(&Url::from_str("https://example.org:443/blah#hahah").unwrap()).as_str(),
            "https://example.org/blah",
        );

        assert_eq!(
            clean_url(&Url::from_str("https://example.org").unwrap()).as_str(),
            "https://example.org/",
        );
    }
}
